import asyncio
import collections.abc
import datetime
import gc
import io
import logging
import math
import os
import shutil
import typing
from asyncio import StreamReader
from random import random
from typing import Optional

import aiofiles
import fsspec
import grpc
from aiofiles.threadpool.binary import AsyncBufferedReader
from flytekit.core.utils import timeit
from fsspec.asyn import AbstractAsyncStreamedFile, AsyncFileSystem, get_loop, sync
from fsspec.callbacks import _DEFAULT_CALLBACK
from fsspec.spec import AbstractBufferedFile
from google.protobuf.duration_pb2 import Duration

from union._async import patch_exception_handler
from union._config import _get_config_obj
from union.filesystems._async_utils import (
    create_channel_with_loop,
    if_coro,
)
from union.filesystems._endpoint import _create_channel
from union.internal.common import list_pb2
from union.internal.objectstore.definition_pb2 import (
    Key,
    Metadata,
    Object,
)
from union.internal.objectstore.objectstore_service_pb2_grpc import (
    ObjectStoreServiceStub,
)
from union.internal.objectstore.payload_pb2 import (
    DeleteRequest,
    DownloadPartRequest,
    GetRequest,
    HeadRequest,
    HeadResponse,
    ListRequest,
    ListResponse,
    MetadataResponse,
    PresignRequest,
    PutRequest,
    StartMultipartUploadRequest,
    SuccessfulUploadRequest,
    TerminateMultipartUploadRequest,
    UploadPartRequest,
    UploadPartResponse,
)

_DEFAULT_LOGGER = logging.getLogger(__name__)

_FS_PREFIX = "union://"

# Shorter default protocol for unionfs
_FS_PREFIX_SHORT = "ufs://"


def _prefix_path(path: str) -> str:
    if path.startswith(_FS_PREFIX):
        return path

    if path.startswith(_FS_PREFIX_SHORT):
        return path
    return f"{_FS_PREFIX}{path.rstrip('/')}"


async def _presigned_download(
    client: ObjectStoreServiceStub,
    rpath: str,
    start: Optional[int] = None,
    end: Optional[int] = None,
) -> bytes:
    """Download and return bytes from rpath."""
    expires_in_pb = Duration()
    expires_in_pb.FromTimedelta(datetime.timedelta(seconds=3 * 60))
    req = PresignRequest(key=Key(key=rpath), expires_in=expires_in_pb)
    presign_response = await client.Presign(req)
    fs = fsspec.filesystem("https", asynchronous=True)

    headers = {}
    if start is not None and end is not None:
        headers["Range"] = f"bytes={start}-{end-1}"

    opened_fs = await fs.open_async(presign_response.signed_url, "rb")
    # Bug in AsyncStreamFile where kwargs is not set correctly
    opened_fs.kwargs = {"headers": headers}
    async with opened_fs as f:
        return await f.read()


async def _fetch_range(client: ObjectStoreServiceStub, max_attempts: int, path: str, start: int, end: int) -> bytes:
    out = io.BytesIO()
    attempt = 0
    path = _prefix_path(path)
    while True:
        try:
            async for resp in client.DownloadPart(
                DownloadPartRequest(key=Key(key=path), start_pos=start, size_bytes=end - start)
            ):
                chunk_contents = resp.object.contents
                out.write(chunk_contents)
            return out.getvalue()
        except grpc.RpcError as e:
            if e.Code() == grpc.StatusCode.RESOURCE_EXHAUSTED:
                attempt += 1
            last_exception = e

        if attempt > max_attempts:
            raise last_exception
        else:
            # TODO(haytham): Make configurable
            await asyncio.sleep(math.pow(2, attempt) * 10 * (random() + 0.5))  # exponential backoff of 10s to 30s


class AsyncUnionFS(AsyncFileSystem):
    mirror_sync_methods = True
    async_impl = True
    cachable = False
    _metadata: Optional[MetadataResponse] = None

    def __init__(
        self,
        logger: logging.Logger = _DEFAULT_LOGGER,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        *args,
        **kwargs,
    ):
        cfg = _get_config_obj(default_to_union_semantics=True)
        if loop is None:
            loop = get_loop()
            remote_connection = create_channel_with_loop(loop, _create_channel, config=cfg)
        else:
            if not loop.is_running():
                raise ValueError("If a loop is passed in, it must be running")
            remote_connection = _create_channel(config=cfg)
        patch_exception_handler(loop)

        self._client = ObjectStoreServiceStub(remote_connection.channel)
        self._in_execution_env = remote_connection.in_execution_env
        if kwargs.get("batch_size") is None:
            batch_size = self.max_concurrent_tasks
        else:
            batch_size = kwargs["batch_size"]
        super().__init__(loop=loop, batch_size=batch_size, *args, **kwargs)
        self._metadata = None
        self.blocksize = 2**18  # 2MB
        self._logger = logger

    @property
    def retries(self) -> int:
        return 3

    @property
    def max_concurrent_tasks(self) -> int:
        return 10

    @property
    def max_attempts(self) -> int:
        return 3

    async def _ensure_metadata(self) -> MetadataResponse:
        if self._metadata is not None:
            return self._metadata

        self._metadata = await self._client.Metadata(Metadata())
        return self._metadata

    async def _rm_file(self, path: str, **kwargs):
        path = _prefix_path(path)
        self._logger.debug(f"Deleting {path}")
        self._client.Delete(DeleteRequest(key=Key(key=path)))

    async def _cp_file(self, path1: str, path2: str, **kwargs):
        raise NotImplementedError("cp_file is not supported for UnionFS")

    async def _get_file(self, rpath: str, lpath: str, callback=_DEFAULT_CALLBACK, **kwargs):
        if os.path.isdir(lpath):
            return
        rpath = _prefix_path(rpath)
        self._logger.debug(f"Getting file async: {rpath}")
        if not self._in_execution_env:
            content = await _presigned_download(self._client, rpath)
            async with aiofiles.open(lpath, "wb") as f0:
                await f0.write(content)

            return

        metadata = await self._ensure_metadata()

        head_resp = await self._client.Head(HeadRequest(key=Key(key=rpath)))
        callback.set_size(head_resp.size_bytes)

        if head_resp.size_bytes <= metadata.max_single_part_object_size_bytes:
            # Download the whole file
            resp = await self._client.Get(GetRequest(key=Key(key=rpath)))
            async with aiofiles.open(lpath, "wb") as f0:
                await f0.write(resp.object.contents)
            callback.relative_update(head_resp.size_bytes)
            return

        # Directory to store downloaded chunks
        output_directory = f"{lpath}.chunks"

        # Create the output directory if it doesn't exist
        os.makedirs(output_directory, exist_ok=True)

        # num_chunks is the number of chunks that the file is split into. It'll be used to name the files
        num_chunks = math.ceil(head_resp.size_bytes / metadata.max_part_size_bytes)

        semaphore = asyncio.Semaphore(self.max_concurrent_tasks)

        async def _download_chunks(start_pos: int = 0, size_bytes: int = -1) -> collections.abc.AsyncIterable[bytes]:
            async for resp in self._client.DownloadPart(
                DownloadPartRequest(key=Key(key=rpath), start_pos=start_pos, size_bytes=size_bytes)
            ):
                chunk_contents = resp.object.contents
                yield chunk_contents
                callback.relative_update(len(chunk_contents))

        async def _download_part(part_number: int = 0):
            file_path = os.path.join(output_directory, f"chunk_{part_number}.dat")
            async with aiofiles.open(file_path, "wb") as local_file:
                async with semaphore:
                    async for chunk in _download_chunks(
                        start_pos=part_number * metadata.max_part_size_bytes,
                        size_bytes=metadata.max_part_size_bytes,
                    ):
                        await local_file.write(chunk)

        # Download and save each chunk to a separate file
        out = []
        for chunk_id in range(0, num_chunks):
            out.append(_download_part(part_number=chunk_id))

        await asyncio.gather(*out)

        # Merge the downloaded chunks into a single file
        async with aiofiles.open(lpath, "wb") as merged_file:
            for chunk_id in range(0, num_chunks):
                file_path = os.path.join(output_directory, f"chunk_{chunk_id}.dat")
                async with aiofiles.open(file_path, "rb") as chunk_file:
                    await merged_file.write(await chunk_file.read())

        # TODO: Delete the downloaded chunks
        # TODO: Logging
        self._logger.debug(f"Merged file saved: {lpath}")
        shutil.rmtree(output_directory)

    async def _cat_file(
        self,
        path: str,
        start: Optional[int] = None,
        end: Optional[int] = None,
        **kwargs,
    ):
        path = _prefix_path(path)
        metadata = await self._ensure_metadata()

        head_resp = await self._client.Head(HeadRequest(key=Key(key=path)))
        if head_resp.size_bytes <= metadata.max_single_part_object_size_bytes:
            # Download the whole file
            resp = await self._client.Get(GetRequest(key=Key(key=path)))
            return resp.object.contents

        if start is None and end is None:
            # Download the whole file
            # TODO: Check file size to decide if we should do multipart download
            resp = await self._client.Get(GetRequest(key=Key(key=path)))
            return resp.object.contents

        raise "Not implemented"

    async def _pipe_file(self, path: str, data: bytes, callback=_DEFAULT_CALLBACK, **kwargs):
        path = _prefix_path(path)
        ret = await self._upload_data(path, io.BytesIO(data), size=len(data), callback=callback)
        gc.collect()
        return ret

    @timeit("Upload a single part")
    async def _upload_chunk(
        self,
        operation_id: str,
        rpath: str,
        chunk: typing.Union[io.BytesIO, StreamReader, AsyncBufferedReader],
        max_upload_size: int,
        part_number: int,
        callback=_DEFAULT_CALLBACK,
    ) -> (UploadPartResponse, bool):
        rpath = _prefix_path(rpath)
        while True:
            attempt = 0
            try:
                result = await self._client.UploadPart(
                    upload_part_request_iter(
                        operation_id=operation_id,
                        key=rpath,
                        buffer=chunk,
                        block_size=self.blocksize,
                        max_upload_size=max_upload_size,
                        logger=self._logger,
                        part_number=part_number,
                        callback=callback,
                    )
                )
                return result, False
            except grpc.RpcError as e:
                self._logger.warning(f"UploadPart: Failed [{e.code()}]. Error: {e.details()}")
                attempt += 1
                last_exception = e
            except EOFError:
                self._logger.warning("UploadPart: EOFError. Returning...")
                return None, True
            except Exception:
                return None, True
            except asyncio.CancelledError:
                return None, True

            if attempt > self.max_attempts:
                self._logger.error("UploadPart: Max attempts reached. Raising exception...")
                raise last_exception
            else:
                # TODO(haytham): Make configurable
                sleep_time = math.pow(2, attempt) * 10 * (random() + 0.5)
                self._logger.warning(f"UploadPart: Retrying in {sleep_time} seconds...")
                await asyncio.sleep(sleep_time)  # exponential backoff of 10s to 30s

    async def _upload_data(
        self,
        path: str,
        data: typing.Union[io.BytesIO, StreamReader, AsyncBufferedReader],
        size: int,
        callback=_DEFAULT_CALLBACK,
    ):
        path = _prefix_path(path)
        metadata = await self._ensure_metadata()

        if size <= metadata.max_single_part_object_size_bytes:
            self._logger.debug("Uploading a single part")
            await self._client.Put(
                PutRequest(
                    key=Key(key=path),
                    object=Object(contents=await if_coro(data.read())),
                )
            )
            callback.relative_update(size)
            return

        self._logger.debug("Uploading a large object")
        resp = await self._client.StartMultipartUpload(
            StartMultipartUploadRequest(
                key=Key(key=path),
                metadata=Metadata(),
            )
        )

        out = []
        part_number = 0

        async def upload_chunks():
            nonlocal part_number
            while True:
                part_number += 1
                self._logger.debug(f"Uploading {path}: part {part_number}")
                # Do not await here, just append the task to the list and we will await all of them later
                upload_task, done = await self._upload_chunk(
                    operation_id=resp.operation_id,
                    rpath=path,
                    chunk=data,
                    max_upload_size=metadata.max_part_size_bytes,
                    part_number=part_number,
                    callback=callback,
                )
                if done:
                    break
                out.append(upload_task)

            # out.extend(await asyncio.gather(*tasks))  # Wait for all upload tasks to complete

        await upload_chunks()

        # part numbers are 1-based
        parts = {o.etag: i + 1 for i, o in enumerate(out)}

        if len(parts) == 0:
            if size > 0:
                raise ValueError(f"No parts uploaded, but size is not zero [{size}]")
            self._logger.debug("No parts uploaded, returning")
            return

        await self._client.TerminateMultipartUpload(
            TerminateMultipartUploadRequest(
                operation_id=resp.operation_id,
                key=Key(key=path),
                successful_upload=SuccessfulUploadRequest(
                    etags_parts=parts,
                ),
            )
        )

    async def _put_file(self, lpath: str, rpath: str, callback=_DEFAULT_CALLBACK, **kwargs):
        if os.path.isdir(lpath):
            raise ValueError("Can't upload a directory")

        rpath = _prefix_path(rpath)
        size = os.path.getsize(lpath)
        callback.set_size(size)
        self._logger.debug(f"Uploading file: {lpath}")
        async with aiofiles.open(lpath, "rb") as f0:
            ret = await self._upload_data(path=rpath, data=f0, size=size, callback=callback)
            gc.collect()
            return ret

    async def _info(self, path: str, **kwargs):
        path = _prefix_path(path)
        req = HeadRequest(key=Key(key=path))
        try:
            res: HeadResponse = await self._client.Head(req)
        except grpc.RpcError as e:
            if e.code() == grpc.StatusCode.NOT_FOUND:
                # If there is no key at that location, check if it's a prefix
                nested_keys = await self._ls(path=path, detail=False)
                self._logger.debug(f"Nested Keys: {len(nested_keys)}")
                if len(nested_keys) > 0:
                    return {
                        "name": path,
                        "type": "directory",
                        "size": 0,
                        "StorageClass": "DIRECTORY",
                    }
                raise FileNotFoundError(path)
            raise
        return {
            "size": res.size_bytes,
            "etag": res.etag,
            "tags": res.metadata.tag if res.metadata and res.metadata.tag else {},
            "type": "file",
        }

    async def _ls(self, path: str, detail: bool = True, **kwargs):
        self._logger.debug(f"Listing: {path}")
        has_prefix = path.startswith(_FS_PREFIX)
        path = _prefix_path(path)
        if path.endswith("*"):
            path = path.strip("*")

        if not path.endswith("/"):
            # Ensure we have a trailing prefix as in fsspec/gcsfs
            # https://github.com/fsspec/gcsfs/blob/5cb4479b0ac9737ed99b906befc88c9250cbd21c/gcsfs/core.py#L611
            path = path + "/"

        req = ListRequest(
            request=list_pb2.ListRequest(
                filters=[
                    list_pb2.Filter(
                        field="prefix",
                        function=list_pb2.Filter.GREATER_THAN_OR_EQUAL,
                        values=[path],
                    )
                ],
            )
        )

        items = []
        while True:
            res: ListResponse = await self._client.List(req)
            for k in res.keys:
                items.append(
                    {
                        "name": k.key if has_prefix else k.key[len(_FS_PREFIX) :],
                        "type": "file",
                        # The API doesn't return information about the object sizes for now.
                        "size": 0,
                    }
                )
            for d in res.directories:
                items.append(
                    {
                        "name": d.key if has_prefix else d.key[len(_FS_PREFIX) :],
                        "type": "directory",
                        # The API doesn't return information about the object sizes for now.
                        "size": 0,
                    }
                )

            if not res.next_token or res.next_token == "":
                break
            req.request.token = res.next_token
        self._logger.debug(f"Items from ls: {items}")
        return items

    async def open_async(self, path: str, mode="rb", **kwargs) -> AbstractAsyncStreamedFile:
        if "b" not in mode or kwargs.get("compression"):
            raise ValueError
        path = _prefix_path(path)
        self._logger.debug(f"Opening file async: {path}")
        m = await self._ensure_metadata()
        return AsyncUnionStreamFile(self, self._client, path, mode, block_size=m.max_part_size_bytes, **kwargs)

    def _open(
        self,
        path,
        mode="rb",
        block_size=None,
        autocommit=True,
        cache_options=None,
        **kwargs,
    ):
        if "b" not in mode or kwargs.get("compression"):
            raise ValueError
        path = _prefix_path(path)
        self._logger.debug(f"Opening file: {path}")
        m = sync(self.loop, self._ensure_metadata)
        return UnionStreamFile(
            self._client,
            fs=self,
            path=path,
            mode=mode,
            block_size=m.max_part_size_bytes,
            autocommit=autocommit,
            cache_options=cache_options,
            **kwargs,
        )


async def upload_part_request_iter(
    logger: logging.Logger,
    operation_id: str,
    key: str,
    buffer: typing.Union[StreamReader, AsyncBufferedReader],
    part_number: int,
    max_upload_size: int,
    callback=_DEFAULT_CALLBACK,
    block_size: int = 2**24,  # 16MB
) -> UploadPartRequest:
    total_read = 0
    while total_read < max_upload_size:
        raw = await if_coro(buffer.read(block_size))
        if not raw or len(raw) == 0:
            if total_read == 0:
                raise EOFError("No data to upload")
            return
        callback.relative_update(len(raw))
        req = UploadPartRequest(
            operation_id=operation_id,
            key=Key(key=key),
            part_number=part_number,
            object=Object(contents=raw),
        )
        yield req
        total_read += len(raw)
    logger.debug(f"Uploaded: {total_read}")


class AsyncUnionStreamFile(AbstractAsyncStreamedFile):
    """
    AsyncUnionStreamFile implements the AbstractAsyncStreamedFile interface for the UnionFS.
    It can be used to read and write files using async.io apis.
    """

    def __init__(
        self,
        fs: AsyncUnionFS,
        object_store_client: ObjectStoreServiceStub,
        path: str,
        mode: str,
        **kwargs,
    ):
        super().__init__(fs, path, mode, **kwargs)
        self._client = object_store_client
        self._part_number = 1

    async def _fetch_range(self, start: int, end: int) -> bytes:
        return await _fetch_range(self._client, self.fs.max_attempts, _prefix_path(self.path), start, end)

    async def _initiate_upload(self):
        resp = await self._client.StartMultipartUpload(
            StartMultipartUploadRequest(key=Key(key=_prefix_path(self.path)), metadata=None)
        )

        self._operation_id = resp.operation_id

    async def _upload_chunk(self, final: bool = False):
        metadata = await self.fs._ensure_metadata()
        await self._client.UploadPart(
            upload_part_request_iter(
                operation_id=self._operation_id,
                logger=self.fs._logger,
                key=_prefix_path(self.path),
                buffer=self.buffer,
                max_upload_size=metadata.max_part_size_bytes,
                block_size=self.blocksize,
                part_number=self._part_number,
            )
        )
        self._part_number = self._part_number + 1
        return True


class UnionStreamFile(AbstractBufferedFile):
    def __init__(
        self,
        client: ObjectStoreServiceStub,
        fs: AsyncUnionFS,
        path,
        mode="rb",
        block_size="default",
        autocommit=True,
        cache_type="readahead",
        cache_options=None,
        size=None,
        **kwargs,
    ):
        super().__init__(
            fs,
            path,
            mode,
            block_size,
            autocommit,
            cache_type,
            cache_options,
            size,
            **kwargs,
        )
        self._client = client
        self._part_number = 1
        self._parts = {}
        self._operation_id = ""
        self._upload_initiated = False

    def _upload_chunk(self, final=False):
        async def internal():
            metadata = await self.fs._ensure_metadata()
            # buffer is at the last byte written, we need to seek to 0 to start uploading.
            await if_coro(self.buffer.seek(0))
            return await self._client.UploadPart(
                upload_part_request_iter(
                    operation_id=self._operation_id,
                    logger=self.fs._logger,
                    key=_prefix_path(self.path),
                    buffer=self.buffer,
                    max_upload_size=metadata.max_part_size_bytes,
                    block_size=self.blocksize,
                    part_number=self._part_number,
                )
            )

        self.fs._logger.debug(f"Uploading chunk {self._part_number}")
        resp = sync(self.fs.loop, internal)
        self._parts[resp.etag] = self._part_number
        self._part_number = self._part_number + 1
        return True

    def _initiate_upload(self):
        self.fs._logger.debug(f"Initiating upload {self.path}")
        resp = sync(
            self.fs.loop,
            self._client.StartMultipartUpload,
            StartMultipartUploadRequest(key=Key(key=_prefix_path(self.path)), metadata=None),
        )

        self._operation_id = resp.operation_id
        self._upload_initiated = True

    def _fetch_range(self, start, end):
        if not self.fs._in_execution_env:
            return sync(
                self.fs.loop,
                _presigned_download,
                self._client,
                _prefix_path(self.path),
                start,
                end,
            )
        return sync(
            self.fs.loop,
            _fetch_range,
            self._client,
            self.fs.max_attempts,
            _prefix_path(self.path),
            start,
            end,
        )

    def close(self):
        super().close()
        if self._upload_initiated:
            return sync(
                self.fs.loop,
                self._client.TerminateMultipartUpload,
                TerminateMultipartUploadRequest(
                    operation_id=self._operation_id,
                    key=Key(key=_prefix_path(self.path)),
                    successful_upload=SuccessfulUploadRequest(
                        etags_parts=self._parts,
                    ),
                ),
            )


fsspec.register_implementation(name="union", cls=AsyncUnionFS, clobber=True)
# Shorter default protocol for unionfs
fsspec.register_implementation(name="ufs", cls=AsyncUnionFS, clobber=True)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: imagebuilder/service.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from union.internal.imagebuilder import payload_pb2 as imagebuilder_dot_payload__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1aimagebuilder/service.proto\x12\x15\x63loudidl.imagebuilder\x1a\x1aimagebuilder/payload.proto2p\n\x0cImageService\x12`\n\x08GetImage\x12&.cloudidl.imagebuilder.GetImageRequest\x1a\'.cloudidl.imagebuilder.GetImageResponse\"\x03\x90\x02\x01\x42\xd1\x01\n\x19\x63om.cloudidl.imagebuilderB\x0cServiceProtoH\x02P\x01Z/github.com/unionai/cloud/gen/pb-go/imagebuilder\xa2\x02\x03\x43IX\xaa\x02\x15\x43loudidl.Imagebuilder\xca\x02\x15\x43loudidl\\Imagebuilder\xe2\x02!Cloudidl\\Imagebuilder\\GPBMetadata\xea\x02\x16\x43loudidl::Imagebuilderb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'imagebuilder.service_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\031com.cloudidl.imagebuilderB\014ServiceProtoH\002P\001Z/github.com/unionai/cloud/gen/pb-go/imagebuilder\242\002\003CIX\252\002\025Cloudidl.Imagebuilder\312\002\025Cloudidl\\Imagebuilder\342\002!Cloudidl\\Imagebuilder\\GPBMetadata\352\002\026Cloudidl::Imagebuilder'
  _IMAGESERVICE.methods_by_name['GetImage']._options = None
  _IMAGESERVICE.methods_by_name['GetImage']._serialized_options = b'\220\002\001'
  _globals['_IMAGESERVICE']._serialized_start=81
  _globals['_IMAGESERVICE']._serialized_end=193
# @@protoc_insertion_point(module_scope)



import glob
import os
import platform
import re
import shlex
from dataclasses import dataclass, field
from datetime import timedelta
from enum import Enum
from typing import Dict, List, Optional, Union

from flyteidl.core.literals_pb2 import KeyValuePair
from flyteidl.core.tasks_pb2 import Container, ContainerPort
from flyteidl.core.tasks_pb2 import Resources as ResourcesIDL
from flytekit import FlyteContextManager, ImageSpec, Resources
from flytekit.core.artifact import ArtifactQuery
from mashumaro.codecs.json import JSONEncoder

from union.app._common import patch_get_flytekit_for_pypi, patch_get_unionai_for_pypi
from union.internal.app.app_definition_pb2 import App as AppIDL
from union.internal.app.app_definition_pb2 import (
    AutoscalingConfig,
    Identifier,
    IngressConfig,
    Meta,
    Replicas,
    Spec,
)
from union.internal.app.app_definition_pb2 import Concurrency as ConcurrencyIDL
from union.internal.app.app_definition_pb2 import RequestRate as RequestRateIDL
from union.internal.app.app_definition_pb2 import ScalingMetric as ScalingMetricIDL


@dataclass
class AppRegistry:
    """Keeps track of all user defined applications"""

    apps: Dict[str, "App"] = field(default_factory=dict)

    def add_app(self, app: "App"):
        execution_state = FlyteContextManager.current_context().execution_state
        if execution_state.mode is None and app.name in self.apps:
            # This is running in a non execute context, so we check if the app name is already used:
            msg = f"App with name: {app.name} was already defined, please use a different name"
            raise ValueError(msg)

        self.apps[app.name] = app


APP_REGISTRY = AppRegistry()
# TODO: Add more protocols here when we deploy to another cloud providers
SUPPORTED_FS_PROTOCOLS = ["s3://", "s3a://", "gs://"]


@dataclass
class URLQuery:
    name: str
    public: bool


ENV_NAME_RE = re.compile("^[_a-zA-Z][_a-zA-Z0-9]*$")
APP_NAME_RE = re.compile(r"[a-z0-9]([-a-z0-9]*[a-z0-9])?(\\.[a-z0-9]([-a-z0-9]*[a-z0-9])?)*")


def _has_file_extension(file_path) -> bool:
    _, ext = os.path.splitext(file_path)
    return bool(ext)


@dataclass
class Input:
    """
    Input for application.

    :param name: Name of input.
    :param value: Value for input.
    :param env_name: Environment name to set the value in the serving environment.
    :param download: When True, the input will be automatically downloaded. This
        only works if the value refers to an item in a object store. i.e. `s3://...`
    """

    class Type(Enum):
        File = "file"
        Directory = "directory"
        String = "string"
        _UrlQuery = "url_query"  # Private type, users should not need this

    name: str
    value: Union[str, ArtifactQuery, URLQuery]
    env_name: Optional[str] = None
    type: Optional[Type] = None
    download: bool = False
    mount: Optional[str] = None

    def __post_init__(self):
        if self.env_name is not None and ENV_NAME_RE.match(self.env_name) is None:
            msg = f"env_name ({self.env_name}) is not a valid environment name for shells"
            raise ValueError(msg)


@dataclass
class MaterializedInput:
    value: str
    type: Optional[Input.Type] = None


@dataclass
class AppSerializationSettings:
    """Runtime settings for creating an AppIDL"""

    org: str
    project: str
    domain: str

    desired_state: Spec.DesiredState
    materialized_inputs: dict[str, MaterializedInput] = field(default_factory=dict)
    additional_distribution: Optional[str] = None


@dataclass
class InputBackend:
    """
    Input information for the backend.
    """

    name: str
    value: str
    download: bool
    type: Optional[Input.Type] = None
    env_name: Optional[str] = None
    dest: Optional[str] = None

    def __post_init__(self):
        if self.type is None:
            if any(self.value.startswith(proto) for proto in SUPPORTED_FS_PROTOCOLS):
                if _has_file_extension(self.value):
                    self.type = Input.Type.File
                else:
                    self.type = Input.Type.Directory
            else:
                self.type = Input.Type.String

        if self.type == Input.Type.String:
            # Nothing to download for a string
            self.download = False

        # If the type is a file or directory and there is a destination, then we
        # automatically assume it is going to be downloaded
        # TODO: In the future, we may mount this, so there is no need to download it
        # with the runtime.
        if self.type in (Input.Type.File, Input.Type.Directory) and self.dest is not None:
            self.download = True

    @classmethod
    def from_input(cls, user_input: Input, settings: AppSerializationSettings) -> "InputBackend":
        if isinstance(user_input.value, str):
            value = user_input.value
            input_type = user_input.type
        else:
            # ArtifactQuery or URLQuery
            try:
                materialized_input = settings.materialized_inputs[user_input.name]
                value = materialized_input.value
                input_type = materialized_input.type or user_input.type
            except KeyError:
                msg = f"Did not materialize {user_input.name}"
                raise ValueError(msg)

        return InputBackend(
            name=user_input.name,
            value=value,
            download=user_input.download,
            env_name=user_input.env_name,
            type=input_type,
            dest=user_input.mount,
        )


@dataclass
class ServeConfig:
    """
    Configuration for serve runtime.

    :param code_uri: Location of user code in an object store (s3://...)
    :param user_inputs: User inputs. Passed in by `app.inputs`
    """

    code_uri: str  # location of user code
    inputs: List[InputBackend]


SERVE_CONFIG_ENCODER = JSONEncoder(ServeConfig)


@dataclass
class ResolvedInclude:
    src: str
    dest: str


class ScalingMetric:
    @dataclass(frozen=True)
    class Concurrency:
        """
        Use this to specify the concurrency metric for autoscaling, i.e. the number of concurrent requests at a replica
         at which to scale up.
        """

        val: int

        def __post_init__(self):
            if self.val < 1:
                raise ValueError("Concurrency must be greater than or equal to 1")

        def _to_union_idl(self) -> ConcurrencyIDL:
            return ConcurrencyIDL(target_value=self.val)

    @dataclass
    class RequestRate:
        """
        Use this to specify the request rate metric for autoscaling, i.e. the number of requests per second at a replica
         at which to scale up.
        """

        val: int

        def __post_init__(self):
            if self.val < 1:
                raise ValueError("Request rate must be greater than or equal to 1")

        def _to_union_idl(self) -> RequestRateIDL:
            return RequestRateIDL(target_value=self.val)

    @staticmethod
    def _to_union_idl(metric: Optional[Union[RequestRate, Concurrency]]) -> Optional[ScalingMetricIDL]:
        if metric is None:
            return None
        request_rate = None
        concurrency = None
        if isinstance(metric, ScalingMetric.RequestRate):
            request_rate = metric._to_union_idl()
        if isinstance(metric, ScalingMetric.Concurrency):
            concurrency = metric._to_union_idl()
        return ScalingMetricIDL(
            request_rate=request_rate,
            concurrency=concurrency,
        )

    @staticmethod
    def _validate(metric: Optional[Union[RequestRate, Concurrency]]):
        if metric is None:
            return
        if not isinstance(metric, (ScalingMetric.RequestRate, ScalingMetric.Concurrency)):
            raise ValueError("scaling_metric must be an instance of Concurrency or RequestRate")


@dataclass
class App:
    """
    App specification.

    :param name: The name of the application.
    :param container_image: The container image to use for the application.
    :param command: Command to start application.
    :param port: Port application listens to. Currently, this must be 8080 and the application
        must listen on 8080.
    :param requests: Compute resource requests for application.
    :param limits: Compute resource limits for application.
    :param include: Files to include for your application.
    :param inputs: Inputs for the application.
    :param env: Environment variables to set for the application.
    :param min_replicas: Minimum number of replicas (ignore if autoscaling is set).
    :param max_replicas: Maximum number of replicas (ignore if autoscaling is set).
    :param scaledown_after: Time to wait before scaling down a replica after it has been idle.
    :param scaling_metric: Autoscale based on a parameter, e.g. request rate or concurrency
      (others may be added in the future).
    :param cluster_pool: The target cluster_pool where the app should be deployed.
    """

    name: str
    container_image: Union[str, ImageSpec]
    command: Union[List[str], str]
    port: int
    limits: Resources
    requests: Resources = field(default_factory=Resources)
    min_replicas: int = 0
    max_replicas: int = 1
    scaledown_after: Optional[Union[int, timedelta]] = None
    scaling_metric: Optional[Union[ScalingMetric.Concurrency, ScalingMetric.RequestRate]] = None
    include: List[str] = field(default_factory=list)
    inputs: List[Input] = field(default_factory=list)
    env: dict = field(default_factory=dict)
    cluster_pool: str = "default"

    _include_resolved: List[ResolvedInclude] = field(default_factory=list)

    def _validate_autoscaling_config(self):
        if self.min_replicas < 0:
            raise ValueError("min_replicas must be greater than or equal to 0")
        if self.max_replicas < 1 or self.max_replicas < self.min_replicas:
            raise ValueError(f"max_replicas must be greater than or equal to 1 or min_replicas:{self.min_replicas}")
        if self.scaledown_after is not None:
            if isinstance(self.scaledown_after, int):
                self.scaledown_after = timedelta(seconds=self.scaledown_after)
            if self.scaledown_after.total_seconds() < 0:
                raise ValueError("scaledown_after must be greater than or equal to 0")
        ScalingMetric._validate(self.scaling_metric)

    def __post_init__(self):
        self._validate_autoscaling_config()
        match_re = APP_NAME_RE.fullmatch(self.name)
        if not match_re:
            raise ValueError("App name must consist of lower case alphanumeric characters, '-' or '.'")

        APP_REGISTRY.add_app(self)

    def _resolve_include(self, app_directory: str, cwd: str) -> "App":
        """Resolve include based on the working_dir.

        If a path in `include` is prefixed with "./", then those files are
        assumed to be relative to the file that has the App object.
        """
        relative_prefix = "./"
        seen_dests = set()

        included_resolved = []

        for file in self.include:
            normed_file = os.path.normpath(file)
            if file.startswith(relative_prefix):
                # File is relative to the app_directory:
                src_dir = app_directory
            else:
                src_dir = cwd

            src = os.path.join(src_dir, normed_file)

            if "*" in src:
                new_srcs = glob.glob(src, recursive=True)
            else:
                new_srcs = [src]

            for new_src in new_srcs:
                dest = os.path.relpath(new_src, src_dir)
                if platform.system() == "Windows":
                    # The runtime is linux so the destinations need to be understood by linux.
                    dest = dest.replace(os.sep, "/")

                if dest in seen_dests:
                    msg = f"{dest} is in include multiple times. Please remove one of them."
                    raise ValueError(msg)

                seen_dests.add(dest)

                included_resolved.append(ResolvedInclude(src=new_src, dest=dest))

        self._include_resolved = included_resolved
        return self

    def _get_image(self) -> str:
        if isinstance(self.container_image, str):
            return self.container_image

        # Build for image Spec
        from flytekit.image_spec.image_spec import ImageBuildEngine

        # TODO: Remove this patching
        # Do not install flytekit or union when building image
        with patch_get_flytekit_for_pypi(), patch_get_unionai_for_pypi():
            ImageBuildEngine.build(self.container_image)
        return self.container_image.image_name()

    def _get_command(self, settings: AppSerializationSettings) -> List[str]:
        args = ["union-serve"]

        serve_config = ServeConfig(
            code_uri=settings.additional_distribution,
            inputs=[InputBackend.from_input(user_input, settings) for user_input in self.inputs],
        )
        args.extend(["--config", SERVE_CONFIG_ENCODER.encode(serve_config)])

        if isinstance(self.command, str):
            command = shlex.split(self.command)
        else:
            command = self.command

        return args + ["--"] + command

    def _get_resources(self) -> ResourcesIDL:
        requests = []
        limits = []

        if self.requests.cpu is not None:
            requests.append(
                ResourcesIDL.ResourceEntry(name=ResourcesIDL.ResourceName.CPU, value=str(self.requests.cpu))
            )
        if self.requests.mem is not None:
            requests.append(
                ResourcesIDL.ResourceEntry(name=ResourcesIDL.ResourceName.MEMORY, value=str(self.requests.mem))
            )
        if self.requests.gpu is not None:
            requests.append(
                ResourcesIDL.ResourceEntry(name=ResourcesIDL.ResourceName.GPU, value=str(self.requests.gpu))
            )
        if self.requests.ephemeral_storage is not None:
            requests.append(
                ResourcesIDL.ResourceEntry(
                    name=ResourcesIDL.ResourceName.EPHEMERAL_STORAGE, value=str(self.requests.ephemeral_storage)
                )
            )

        if self.limits.cpu is not None:
            limits.append(ResourcesIDL.ResourceEntry(name=ResourcesIDL.ResourceName.CPU, value=str(self.limits.cpu)))
        if self.limits.mem is not None:
            limits.append(ResourcesIDL.ResourceEntry(name=ResourcesIDL.ResourceName.MEMORY, value=str(self.limits.mem)))
        if self.limits.gpu is not None:
            limits.append(ResourcesIDL.ResourceEntry(name=ResourcesIDL.ResourceName.GPU, value=str(self.limits.gpu)))
        if self.limits.ephemeral_storage is not None:
            limits.append(
                ResourcesIDL.ResourceEntry(
                    name=ResourcesIDL.ResourceName.EPHEMERAL_STORAGE, value=str(self.limits.ephemeral_storage)
                )
            )

        return ResourcesIDL(requests=requests, limits=limits)

    def _get_container(self, settings: AppSerializationSettings) -> Container:
        return Container(
            image=self._get_image(),
            command=[],
            args=self._get_command(settings),
            resources=self._get_resources(),
            ports=[
                ContainerPort(container_port=self.port),
            ],
            env=[KeyValuePair(key=k, value=v) for k, v in self.env.items()],
        )

    def _to_union_idl(self, settings: AppSerializationSettings) -> AppIDL:
        scaling_metric = ScalingMetric._to_union_idl(self.scaling_metric)

        dur = None
        if self.scaledown_after:
            from google.protobuf.duration_pb2 import Duration

            dur = Duration()
            dur.FromTimedelta(self.scaledown_after)

        autoscaling = AutoscalingConfig(
            replicas=Replicas(min=self.min_replicas, max=self.max_replicas),
            scaledown_period=dur,
            scaling_metric=scaling_metric,
        )
        return AppIDL(
            metadata=Meta(
                id=Identifier(
                    org=settings.org,
                    project=settings.project,
                    domain=settings.domain,
                    name=self.name,
                ),
            ),
            spec=Spec(
                container=self._get_container(settings),
                desired_state=settings.desired_state,
                ingress=IngressConfig(private=False),
                autoscaling=autoscaling,
                cluster_pool=self.cluster_pool,
            ),
        )

    def query_endpoint(self, *, public: bool) -> URLQuery:
        """
        Query for endpoint.

        :param public: Whether to return the public or internal endpoint.
        :returns: Object representing a URL query.
        """
        return URLQuery(name=self.name, public=public)


@dataclass
class Endpoint(App):
    """
    Endpoint specification.

    :param name: The name of the application.
    :param container_image: The container image to use for the application.
    :param command: Command to start application.
    :param port: Port application listens to. Currently, this must be 8080 and the application
        must listen on 8080.
    :param requests: Compute resource requests for application.
    :param limits: Compute resource limits for application.
    :param include: Files to include for your application.
    :param inputs: Inputs for the application.
    """



"""Module to manage applications on Union."""

import asyncio
import logging
import os
import tarfile
from copy import deepcopy
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import List, Optional

import grpc
from click import ClickException
from flyteidl.core.tasks_pb2 import Resources
from flytekit.configuration import Config
from flytekit.core.artifact import ArtifactQuery
from flytekit.models.core.types import BlobType
from rich.console import Console

from union._config import (
    ConfigSource,
    ConfigWithSource,
    _get_config_obj,
    _get_organization,
)
from union.app import App
from union.app._models import AppSerializationSettings, Input, MaterializedInput, URLQuery
from union.cli._common import _get_channel_with_org
from union.internal.app.app_definition_pb2 import App as AppIDL
from union.internal.app.app_definition_pb2 import Identifier, Spec, Status
from union.internal.app.app_payload_pb2 import (
    CreateRequest,
    CreateResponse,
    GetRequest,
    GetResponse,
    ListResponse,
    UpdateRequest,
    UpdateResponse,
    WatchRequest,
    WatchResponse,
)
from union.internal.app.app_payload_pb2 import (
    ListRequest as ListAppsRequest,
)
from union.internal.app.app_service_pb2_grpc import AppServiceStub
from union.internal.common.identifier_pb2 import ProjectIdentifier
from union.internal.common.list_pb2 import ListRequest
from union.remote._remote import UnionRemote

logger = logging.getLogger(__name__)

FILES_TAR_FILE_NAME = "include-files.tar.gz"


class AppRemote:
    def __init__(
        self,
        project: str,
        domain: str,
        config: Optional[Config] = None,
    ):
        if config is None:
            config = _get_config_obj(config, default_to_union_semantics=True)
        else:
            config_with_source = ConfigWithSource(config=config, source=ConfigSource.REMOTE)

            config = _get_config_obj(config_with_source, default_to_union_semantics=True)

        self.config = config
        self.project = project
        self.domain = domain
        self._union_remote = UnionRemote(config=config, default_domain=domain, default_project=project)

    def list(self) -> List[AppIDL]:
        def create_list_request(token: str):
            return ListAppsRequest(
                request=ListRequest(token=token),
                org=self.org,
                project=ProjectIdentifier(name=self.project, domain=self.domain, organization=self.org),
            )

        results = []
        response: ListResponse
        token, has_next = "", True

        while has_next:
            list_request = create_list_request(token=token)

            response = self.sync_client.List(list_request)
            token = response.token
            has_next = token != ""

            results.extend(response.apps)

        return results

    def create_or_update(self, app: App):
        try:
            self.get(name=app.name)
        except grpc.RpcError as e:
            if e.code() == grpc.StatusCode.NOT_FOUND:
                self.create(app)
                return
            raise

        try:
            self.update(app)
        except grpc.RpcError as e:
            if e.code() == grpc.StatusCode.ABORTED and "Either the change has already" in e.details():
                console = Console()
                console.print(f"{app.name} app was already deployed with the current spec")
                return
            raise

    def get(self, name: str) -> AppIDL:
        app_id = Identifier(org=self.org, project=self.project, domain=self.domain, name=name)
        get_app_request = GetRequest(app_id=app_id)

        get_response: GetResponse = self.sync_client.Get(get_app_request)
        app = get_response.app

        return app

    def start(self, name: str):
        app_idl = self.get(name)
        app_idl.spec.desired_state = Spec.DesiredState.DESIRED_STATE_STARTED

        def run_request():
            update_request = UpdateRequest(app=app_idl)
            update_response: UpdateResponse = self.sync_client.Update(update_request)
            return update_response.app

        self._watch_update(app_idl, run_request)

    def update(self, app: App):
        additional_distribution = self.upload_files(app)
        materialized_input_values = self.materialize_values(app)

        settings = AppSerializationSettings(
            org=self.org,
            project=self.project,
            domain=self.domain,
            additional_distribution=additional_distribution,
            desired_state=Spec.DesiredState.DESIRED_STATE_STARTED,
            materialized_inputs=materialized_input_values,
        )
        new_app_idl = app._to_union_idl(settings=settings)

        get_app_request = GetRequest(app_id=new_app_idl.metadata.id)
        get_response: GetResponse = self.sync_client.Get(get_app_request)

        old_app_idl = get_response.app

        updated_app_idl = AppIDL(
            metadata=old_app_idl.metadata,
            spec=new_app_idl.spec,
            status=old_app_idl.status,
        )

        def run_request():
            update_request = UpdateRequest(app=updated_app_idl)
            update_response: UpdateResponse = self.sync_client.Update(update_request)
            return update_response.app

        self._watch_update(updated_app_idl, run_request)

    def create(self, app: App):
        additional_distribution = self.upload_files(app)
        materialized_input_values = self.materialize_values(app)

        settings = AppSerializationSettings(
            org=self.org,
            project=self.project,
            domain=self.domain,
            additional_distribution=additional_distribution,
            desired_state=Spec.DesiredState.DESIRED_STATE_STARTED,
            materialized_inputs=materialized_input_values,
        )
        app_idl = app._to_union_idl(settings=settings)

        def run_request():
            create_request = CreateRequest(app=app_idl)
            create_response: CreateResponse = self.sync_client.Create(create_request)
            return create_response.app

        self._watch_update(app_idl, run_request)

    def _watch_update(self, app_idl, run_request):
        console = Console()
        name = app_idl.metadata.id.name

        _console_url = self.generate_console_url(app_idl)
        console_url = f"[link={_console_url}]{_console_url}[/link]"
        console.print(f"✨ Deploying Application: [link={_console_url}]{name}[/link]")
        console.print(f"🔎 Console URL: {console_url}")

        updated_idl = run_request()
        watch_request = WatchRequest(app_id=updated_idl.metadata.id)

        async def watch():
            response: WatchResponse
            latest_message = ""
            async for response in self.async_client.Watch(watch_request):
                status = response.update_event.updated_app.status
                if not status.conditions:
                    continue
                latest_condition = status.conditions[-1]
                deployment_status = latest_condition.deployment_status
                status_str = Status.DeploymentStatus.Name(deployment_status).split("_")[-1].title()

                if latest_condition.message != latest_message:
                    console.print(f"[bold]\\[Status][/] [italic]{status_str}:[/] {latest_condition.message}")
                    latest_message = latest_condition.message

                if deployment_status in (
                    Status.DeploymentStatus.DEPLOYMENT_STATUS_UNSPECIFIED,
                    Status.DeploymentStatus.DEPLOYMENT_STATUS_STOPPED,
                    Status.DeploymentStatus.DEPLOYMENT_STATUS_FAILED,
                ):
                    raise ClickException("Application failed to deploy")
                elif deployment_status == Status.DEPLOYMENT_STATUS_UNASSIGNED:
                    raise ClickException("Application was unassigned")
                elif deployment_status == Status.DeploymentStatus.DEPLOYMENT_STATUS_STARTED:
                    return

        loop = asyncio.get_event_loop()
        loop.run_until_complete(watch())

        _url = updated_idl.status.ingress.public_url
        url = f"[link={_url}]{_url}[/link]"

        console.print()
        console.print(f"🚀 Deployed Endpoint: {url}")

    def stop(self, name: str):
        app_idl = self.get(name)
        app_idl.spec.desired_state = Spec.DesiredState.DESIRED_STATE_STOPPED

        console = Console()

        console_url = self.generate_console_url(app_idl)
        console.print(f"🛑 Stopping Application: [link={console_url}]{name}[/link]")

        update_request = UpdateRequest(app=app_idl)
        self.sync_client.Update(update_request)

    def upload_files(self, app: App) -> Optional[str]:
        """Upload files required by app."""
        if not app.include:
            return None

        with TemporaryDirectory() as temp_dir:
            tar_path = os.path.join(temp_dir, FILES_TAR_FILE_NAME)
            with tarfile.open(tar_path, "w:gz") as tar:
                for resolve_include in app._include_resolved:
                    tar.add(resolve_include.src, arcname=resolve_include.dest)

            _, upload_native_url = self._union_remote.upload_file(Path(tar_path))

            return upload_native_url

    def materialize_values(self, app: App) -> dict:
        output = {}
        for user_input in app.inputs:
            if isinstance(user_input.value, ArtifactQuery):
                query = deepcopy(user_input.value)
                query.project = self.project
                query.domain = self.domain

                result = self._union_remote.get_artifact(query=query.to_flyte_idl())
                scalar = result.literal.scalar

                input_type = None
                if scalar.blob is not None:
                    # Handle FlyteFiles and FlyteDirectories
                    if scalar.blob.metadata.type.dimensionality == BlobType.BlobDimensionality.SINGLE:
                        input_type = Input.Type.File
                    elif scalar.blob.metadata.type.dimensionality == BlobType.BlobDimensionality.MULTIPART:
                        input_type = Input.Type.Directory
                    value = result.literal.scalar.blob.uri
                elif scalar.primitive.string_value is not None:
                    value = scalar.primitive.string_value
                    input_type = Input.Type.String
                else:
                    msg = f"{scalar} is not supported as an App input"
                    raise ValueError(msg)

                output[user_input.name] = MaterializedInput(
                    value=value,
                    type=input_type,
                )

            elif isinstance(user_input.value, URLQuery):
                query = user_input.value
                # TODO: Assuming application has the same project and domain
                # TODO: Raise more informative error the assumption does not hold
                app_idl = self.get(name=query.name)
                if user_input.value.public:
                    output[user_input.name] = MaterializedInput(
                        value=app_idl.status.ingress.public_url,
                        type=Input.Type.String,
                    )
                else:
                    app_id = app_idl.metadata.id
                    output[user_input.name] = MaterializedInput(
                        value=app_id.name,
                        type=Input.Type._UrlQuery,
                    )

        return output

    def generate_console_url(self, app_idl: AppIDL) -> str:
        """Generate console url for app_idl."""
        http_domain = self._union_remote.generate_console_http_domain()
        org = _get_organization(self.config.platform, self.sync_channel)

        app_id = app_idl.metadata.id

        if org is None or org == "":
            console = "console"
        else:
            console = f"org/{org}"

        return f"{http_domain}/{console}/projects/{app_id.project}/domains/{app_id.domain}/apps/{app_id.name}"

    @property
    def sync_channel(self) -> grpc.Channel:
        try:
            return self._sync_channel
        except AttributeError:
            self._sync_channel, self._org = _get_channel_with_org(self.config.platform)
            return self._sync_channel

    @property
    def org(self) -> Optional[str]:
        try:
            return self._org
        except AttributeError:
            self._sync_channel, self._org = _get_channel_with_org(self.config.platform)
            if self._org is None or self._org == "":
                self._org = self.config.platform.endpoint.split(".")[0]
            return self._org

    @property
    def sync_client(self) -> AppServiceStub:
        try:
            return self._sync_client
        except AttributeError:
            self._sync_client = AppServiceStub(self.sync_channel)
            return self._sync_client

    @property
    def async_client(self) -> AppServiceStub:
        try:
            return self._async_client
        except AttributeError:
            self._async_client = AppServiceStub(self.async_channel)
            return self._async_client

    @property
    def async_channel(self) -> grpc.aio.Channel:
        from union.filesystems._endpoint import _create_secure_channel_from_config

        try:
            return self._async_channel
        except AttributeError:
            self._async_channel = _create_secure_channel_from_config(self.config.platform, self.sync_channel)
            return self._async_channel

    @staticmethod
    def deployment_status(app_idl: AppIDL) -> str:
        try:
            current_status = app_idl.status.conditions[-1].deployment_status
            return Status.DeploymentStatus.Name(current_status).split("_")[-1].title()
        except Exception:
            return "Unknown"

    @staticmethod
    def desired_state(app_idl: AppIDL) -> str:
        return Spec.DesiredState.Name(app_idl.spec.desired_state).split("_")[-1].title()

    @staticmethod
    def get_limits(app_idl: AppIDL) -> dict:
        output = {}
        for limit in app_idl.spec.container.resources.limits:
            if limit.name == Resources.ResourceName.CPU:
                output["cpu"] = limit.value
            if limit.name == Resources.ResourceName.MEMORY:
                output["memory"] = limit.value
        return output



import rich_click as click

from union._config import _DEFAULT_DOMAIN, _DEFAULT_PROJECT_BYOC
from union.remote._app_remote import AppRemote


@click.group()
def stop():
    """Stop a resource."""


@stop.command()
@click.option("--name", type=str, required=True)
@click.option("--project", default=_DEFAULT_PROJECT_BYOC, help="Project name")
@click.option("--domain", default=_DEFAULT_DOMAIN, help="Domain name")
def apps(name: str, project: str, domain: str):
    app_remote = AppRemote(project=project, domain=domain)
    app_remote.stop(name=name)



"""All imports to unionai in this file should be in the function definition.

This plugin is loaded by flytekit, so any imports to unionai can lead to circular imports.
"""

import json
import os
import re
import shutil
import sys
import tempfile
import warnings
from dataclasses import fields
from datetime import datetime, timedelta, timezone
from functools import lru_cache
from importlib.metadata import version
from pathlib import Path
from typing import Optional, Tuple

import click
from flytekit import Labels
from flytekit.constants import CopyFileDetection
from flytekit.exceptions.user import FlyteEntityNotExistException
from flytekit.image_spec.image_spec import _F_IMG_ID, ImageBuildEngine, ImageSpec, ImageSpecBuilder
from flytekit.models.core import execution as core_execution_models
from flytekit.models.filters import ValueIn
from flytekit.remote import FlyteRemote, FlyteWorkflowExecution
from flytekit.remote.remote import MOST_RECENT_FIRST
from flytekit.tools.ignore import DockerIgnore, GitIgnore, IgnoreGroup, StandardIgnore
from flytekit.tools.script_mode import ls_files
from flytekit.tools.translator import Options

from union._utils import _sanitize_label

RECOVER_EXECUTION_PAGE_LIMIT = 1_000
UNIONAI_IMAGE_NAME_KEY = "unionai_image_name"

_PACKAGE_NAME_RE = re.compile(r"^[\w-]+")


def _is_union(package: str) -> bool:
    """Return True if `package` is union or unionai."""
    m = _PACKAGE_NAME_RE.match(package)
    if not m:
        return False
    name = m.group()
    return name in ("union", "unionai")


def get_unionai_for_pypi():
    union_version = version("union")
    if "dev" in union_version:
        return "union"
    else:
        return f"union=={union_version}"


@lru_cache
def _get_remote() -> FlyteRemote:
    from union.configuration._plugin import UnionAIPlugin

    remote = UnionAIPlugin.get_remote(
        config=None,
        project="default",
        domain="development",
    )
    return remote


@lru_cache
def _get_fully_qualified_image_name(remote, execution):
    return execution.outputs["fully_qualified_image"]


def _build(spec: Path, context: Optional[Path], target_image: str) -> str:
    """Build image using UnionAI."""

    remote = _get_remote()
    start = datetime.now(timezone.utc)

    context_url = "" if context is None else remote.upload_file(context)[1]
    spec_url = remote.upload_file(spec)[1]
    entity = remote.fetch_task(project="system", domain="production", name="build-image")

    execution = remote.execute(
        entity,
        project="system",
        domain="development",
        inputs={"spec": spec_url, "context": context_url, "target_image": target_image},
        options=Options(
            labels=Labels(values={UNIONAI_IMAGE_NAME_KEY: _sanitize_label(target_image)}),
        ),
        overwrite_cache=True,
    )
    click.secho("👍 Build submitted!", bold=True, fg="yellow")

    console_url = remote.generate_console_url(execution)

    click.secho(
        "⏳ Waiting for build to finish at: " + click.style(console_url, fg="cyan"),
        bold=True,
    )
    execution = remote.wait(execution, poll_interval=timedelta(seconds=1))

    elapsed = str(datetime.now(timezone.utc) - start).split(".")[0]

    if execution.closure.phase == core_execution_models.WorkflowExecutionPhase.SUCCEEDED:
        click.secho(f"✅ Build completed in {elapsed}!", bold=True, fg="green")
    else:
        error_msg = execution.error.message
        raise click.ClickException(
            f"❌ Build failed in {elapsed} at {click.style(console_url, fg='cyan')} with error:\n\n{error_msg}"
        )

    return _get_fully_qualified_image_name(remote, execution)


def _copy_files_into_build_context(image_spec: ImageSpec, spec: dict, context_path: Path):
    if image_spec.source_copy_mode is not None and image_spec.source_copy_mode != CopyFileDetection.NO_COPY:
        if not image_spec.source_root:
            raise ValueError(f"Field source_root for {image_spec} must be set when copy is set")

        # Easter egg
        # Load in additional packages before installing pip/apt packages
        vendor_path = Path(image_spec.source_root) / ".vendor"
        if vendor_path.is_dir():
            spec["dist_dirpath"] = ".vendor"

        ignore = IgnoreGroup(image_spec.source_root, [GitIgnore, DockerIgnore, StandardIgnore])

        ls, _ = ls_files(
            str(image_spec.source_root), image_spec.source_copy_mode, deref_symlinks=False, ignore_group=ignore
        )

        for file_to_copy in ls:
            rel_path = os.path.relpath(file_to_copy, start=str(image_spec.source_root))
            Path(context_path / rel_path).parent.mkdir(parents=True, exist_ok=True)
            shutil.copy(
                file_to_copy,
                context_path / rel_path,
            )

    if image_spec.copy:
        for src in image_spec.copy:
            src_path = Path(src)

            if src_path.is_absolute() or ".." in src_path.parts:
                raise ValueError("Absolute paths or paths with '..' are not allowed in COPY command.")

            dst_path = context_path / src_path
            dst_path.parent.mkdir(parents=True, exist_ok=True)

            if src_path.is_dir():
                shutil.copytree(src_path, dst_path, dirs_exist_ok=True)
            else:
                shutil.copy(src_path, dst_path)


class UCImageSpecBuilder(ImageSpecBuilder):
    """ImageSpec builder for UnionAI."""

    _SUPPORTED_IMAGE_SPEC_PARAMETERS = {
        "name",
        "builder",
        "python_version",
        "source_root",
        "env",
        "packages",
        "requirements",
        "apt_packages",
        "cuda",
        "cudnn",
        "platform",
        "pip_index",
        "pip_extra_index_url",
        "commands",
        "conda_packages",
        "conda_channels",
        "source_copy_mode",
    }

    def build_image(self, image_spec: ImageSpec):
        """Build image using UnionAI."""
        image_name = image_spec.image_name()

        with tempfile.TemporaryDirectory() as tmpdir:
            tmp_path = Path(tmpdir)
            spec_path, archive_path = self._validate_configuration(image_spec, tmp_path, image_name)
            return _build(spec_path, archive_path, image_name)

    def should_build(self, image_spec: ImageSpec) -> bool:
        """Check whether the image should be built."""
        image_name = image_spec.image_name()
        remote = _get_remote()

        _check_image_builder_enabled(remote)

        try:
            ImageBuildEngine._IMAGE_NAME_TO_REAL_NAME[image_name] = remote._get_image_fqin(image_name)
            click.secho(f"Image {image_name} found. Skip building.", fg="blue")
            return False
        except FlyteEntityNotExistException:
            click.secho(f"Image {image_name} was not found or has expired.", fg="blue")
            click.secho("🐳 Submitting a new build...", bold=True, fg="blue")
            return True

    def _validate_configuration(
        self, image_spec: ImageSpec, tmp_path: Path, image_name: str
    ) -> Tuple[Path, Optional[Path]]:
        """Validate and write configuration for builder."""
        field_names = [f.name for f in fields(ImageSpec) if not f.name.startswith("_")]

        unsupported_parameters = [
            name
            for name in field_names
            if getattr(image_spec, name) and name not in self._SUPPORTED_IMAGE_SPEC_PARAMETERS
        ]
        if unsupported_parameters:
            msg = f"The following parameters are unsupported and ignored: {unsupported_parameters}"
            warnings.warn(msg, UserWarning)

        # Hardcoded for now since our base image only supports 3.11
        if image_spec.python_version is None:
            major = sys.version_info.major
            minor = sys.version_info.minor
        else:
            version_split = image_spec.python_version.split(".")
            if not (2 <= len(version_split) <= 3):
                raise ValueError("python_version must be in the form major.minor. For example: 3.12")
            major, minor = version_split[:2]
            try:
                major, minor = int(major), int(minor)
            except ValueError:
                raise ValueError("python_version must be in the form major.minor. For example: 3.12")

        if not (major == 3 and 9 <= minor <= 12):
            raise ValueError(f"Only python versions between 3.9 to 3.12 are supported, got {major}.{minor}")

        python_version = f"{major}.{minor}"

        spec = {"python_version": python_version}
        # Transform image spec into a spec we expect
        if image_spec.apt_packages:
            spec["apt_packages"] = image_spec.apt_packages
        if image_spec.commands:
            spec["commands"] = image_spec.commands
        if image_spec.cuda or image_spec.cudnn:
            spec["enable_gpu"] = True
        if image_spec.conda_packages:
            spec["conda_packages"] = image_spec.conda_packages
        if image_spec.conda_channels:
            spec["conda_channels"] = image_spec.conda_channels

        env = image_spec.env or {}
        env = {**{_F_IMG_ID: image_name}, **env}
        if env:
            spec["env"] = env

        packages = []
        if image_spec.packages:
            packages.extend(image_spec.packages)

        spec["python_packages"] = packages

        pip_extra_index_urls = []
        if image_spec.pip_index:
            pip_extra_index_urls.append(image_spec.pip_index)

        if image_spec.pip_extra_index_url:
            pip_extra_index_urls.extend(image_spec.pip_extra_index_url)

        spec["pip_extra_index_urls"] = pip_extra_index_urls

        context_path = tmp_path / "build.uc-image-builder"
        context_path.mkdir(exist_ok=True)

        _all_packages = []
        _all_packages.extend(packages)

        if image_spec.requirements:
            if not Path(image_spec.requirements).exists():
                raise ValueError(f"requirements file {image_spec.requirements} does not exist")
            name = Path(image_spec.requirements).name
            shutil.copy2(image_spec.requirements, context_path / name)
            spec["python_requirements_files"] = [name]
            if str(image_spec.requirements).endswith(".txt"):
                with (context_path / name).open() as f:
                    _all_packages.extend([line.strip() for line in f.readlines()])
            if str(image_spec.requirements).endswith("poetry.lock"):
                pyproject = Path(image_spec.requirements).parent.joinpath("pyproject.toml")
                self._check_pyproject_for_unsupported_features(pyproject)
                self._check_image_spec_for_unsupported_features(image_spec)
                if pyproject.exists():
                    # almost never happens, but if the user has 'subdir/poetry.lock', note that this will copy the
                    # pyproject file directly into the context folder, removing the subdir.
                    shutil.copy2(pyproject, context_path / pyproject.name)
                else:
                    raise ValueError(f"must have project file if using lock file {image_spec.requirements}")

        # if union is not specified, then add it here.
        if not any(_is_union(p) for p in _all_packages):
            packages.append(get_unionai_for_pypi())

        _copy_files_into_build_context(image_spec, spec, context_path)

        _copy_files_into_build_context(image_spec, spec, context_path)

        if any(context_path.iterdir()):
            archive_path = Path(shutil.make_archive(tmp_path / "context", "xztar", context_path))
        else:
            archive_path = None

        spec_path = tmp_path / "spec.json"
        with spec_path.open("w") as f:
            json.dump(spec, f)

        return (spec_path, archive_path)

    @staticmethod
    def _check_pyproject_for_unsupported_features(pyproject: Path):
        """
        Use of external sources is prohibited because caching might interfere with other builds, and because the
        image builder might not have network access to it.
        If there's enough request for this for non-pypi but public sources, we can build out smarter caching in all
        our builders.
        See https://python-poetry.org/docs/repositories/#package-sources
        """
        for line in pyproject.read_text().splitlines():
            if "tool.poetry.source" in line:
                raise ValueError("External sources are not supported in pyproject.toml")

    @staticmethod
    def _check_image_spec_for_unsupported_features(image_spec: ImageSpec):
        """
        Use of external sources is prohibited in poetry because it should be declared in the pyproject file instead
        (which in turn should raise an error because we don't want to support non-pypi sources).
        """
        if image_spec.pip_index:
            raise ValueError("External sources are not supported in image spec when using poetry lock file")
        if image_spec.pip_extra_index_url:
            raise ValueError("External sources are not supported in image spec when using poetry lock file")


def _check_image_builder_enabled(remote: FlyteRemote):
    """Checks if the Union image builder is enabled."""
    try:
        remote.fetch_task(project="system", domain="production", name="build-image")
    except FlyteEntityNotExistException as e:
        msg = "Union remote image builder is not enabled. Please contact Union support to enable it."
        raise click.ClickException(msg) from e


def _get_latest_image_build_execution(remote: FlyteRemote, image_name: str) -> Optional[FlyteWorkflowExecution]:
    """
    Get the latest image build execution for the given image name.
    Returns None if no execution is found or the execution is not in QUEUED, RUNNING, or SUCCEEDED phase.
    """
    # Replace ':' with '_' since flyte does not allow ':' in the label value
    target_image = _sanitize_label(image_name)
    executions, _ = remote.client.list_executions_paginated(
        project="system",
        domain="development",
        limit=1,
        filters=[
            ValueIn("execution_tag.key", [UNIONAI_IMAGE_NAME_KEY]),
            ValueIn("execution_tag.value", [target_image]),
            ValueIn("phase", ["QUEUED", "RUNNING", "SUCCEEDED"]),
        ],
        sort_by=MOST_RECENT_FIRST,
    )

    if len(executions) == 0:
        return None

    return remote.sync_execution(FlyteWorkflowExecution.promote_from_model(executions[0]))


def _register_union_image_builder(priority: int = 10):
    ImageBuildEngine.register("unionai", UCImageSpecBuilder(), priority)
    ImageBuildEngine.register("union", UCImageSpecBuilder(), priority)


def _is_output_not_found(e: Exception) -> bool:
    # Unfortunately, failure to find an execution output doesn't return a 404 but an exception that wraps the 404.
    # Here we check the exception string directly
    return "status code: 404" in str(e)



import logging
import os
import typing

from flyteidl.core import artifact_id_pb2 as art_id
from flytekit import BlobType
from flytekit.clients.friendly import SynchronousFlyteClient
from flytekit.configuration import Config
from flytekit.core.type_engine import LiteralsResolver, TypeEngine
from flytekit.exceptions import user as user_exceptions
from flytekit.models.literals import Blob, BlobMetadata, Literal, Scalar
from flytekit.remote import FlyteRemote
from flytekit.remote.entities import FlyteLaunchPlan, FlyteTask, FlyteWorkflow
from flytekit.remote.executions import FlyteNodeExecution, FlyteTaskExecution, FlyteWorkflowExecution
from flytekit.tools.translator import Options

from union._config import (
    _DEFAULT_DOMAIN,
    _DEFAULT_PROJECT_BYOC,
    ConfigSource,
    ConfigWithSource,
    _config_from_api_key,
    _get_config_obj,
    _get_default_project,
    _get_organization,
)
from union._interceptor import update_client_with_interceptor
from union.artifacts import Artifact
from union.internal.artifacts import artifacts_pb2, artifacts_pb2_grpc
from union.internal.imagebuilder import definition_pb2 as image_definition__pb2
from union.internal.imagebuilder import payload_pb2 as image_payload__pb2
from union.internal.imagebuilder import service_pb2_grpc as image_service_pb2_grpc

SERVERLESS_VANITY_URLS = {
    "https://serverless-1.us-east-2.s.union.ai": "https://serverless.union.ai",
    "https://serverless-preview.canary.unionai.cloud": "https://serverless.canary.union.ai",
    "https://serverless-gcp.cloud-staging.union.ai": "https://serverless.staging.union.ai",
}


class UnionRemote(FlyteRemote):
    def __init__(
        self,
        config: typing.Optional[Config] = None,
        default_project: typing.Optional[str] = None,
        default_domain: typing.Optional[str] = _DEFAULT_DOMAIN,
        data_upload_location: str = "flyte://my-s3-bucket/",
        **kwargs,
    ):
        from union._config import _get_image_builder_priority
        from union.ucimage._image_builder import _register_union_image_builder

        if config is None:
            config = _get_config_obj(config, default_to_union_semantics=True)

        config_with_source = ConfigWithSource(config=config, source=ConfigSource.REMOTE)
        config = _get_config_obj(config_with_source, default_to_union_semantics=True)

        if default_project is None:
            default_project = _get_default_project(_DEFAULT_PROJECT_BYOC, cfg_obj=config)

        # register Union image builder when getting remote so it's available for
        # jupyter notebook task and workflow execution.
        _register_union_image_builder(_get_image_builder_priority(config.platform.endpoint))

        super().__init__(config, default_project, default_domain, data_upload_location, **kwargs)
        self._artifacts_client = None
        self._images_client = None

    @classmethod
    def from_api_key(
        cls,
        api_key: str,
        default_project: typing.Optional[str] = None,
        default_domain: typing.Optional[str] = _DEFAULT_DOMAIN,
        data_upload_location: str = "flyte://my-s3-bucket/",
        **kwargs,
    ) -> "UnionRemote":
        """Call this if you want to directly instantiate a UnionRemote from an API key"""
        return cls(
            config=_config_from_api_key(api_key),
            default_project=default_project,
            default_domain=default_domain,
            data_upload_location=data_upload_location,
            **kwargs,
        )

    @property
    def client(self) -> SynchronousFlyteClient:
        """Return a SynchronousFlyteClient for additional operations."""
        if not self._client_initialized:
            client = SynchronousFlyteClient(self.config.platform, **self._kwargs)
            org = _get_organization(self.config.platform, channel=client._channel)
            self._client = update_client_with_interceptor(client, org)
            self._client_initialized = True

        return self._client

    def generate_console_http_domain(self) -> str:
        default_console_http_domain = super().generate_console_http_domain()
        return SERVERLESS_VANITY_URLS.get(default_console_http_domain, default_console_http_domain)

    def generate_console_url(
        self,
        entity: typing.Union[
            FlyteWorkflowExecution,
            FlyteNodeExecution,
            FlyteTaskExecution,
            FlyteWorkflow,
            FlyteTask,
            FlyteLaunchPlan,
            Artifact,
        ],
    ):
        """
        Generate a UnionAI console URL for the given Flyte remote endpoint.
        It will also handle Union AI specific entities like Artifacts.

        This will automatically determine if this is an execution or an entity
        and change the type automatically.
        """
        org = _get_organization(self.config.platform)
        if isinstance(entity, Artifact):
            url = f"{self.generate_console_http_domain()}/console/projects/{entity.project}/domains/{entity.domain}/artifacts/{entity.name}/versions/{entity.version}"  # noqa: E501
        else:
            url = super().generate_console_url(entity)
        if org is None:
            return url

        console_http_domain = self.generate_console_http_domain()
        old_prefix = f"{console_http_domain}/console/"
        new_prefix = f"{console_http_domain}/org/{org}/"

        return url.replace(old_prefix, new_prefix)

    @property
    def artifacts_client(self) -> artifacts_pb2_grpc.ArtifactRegistryStub:
        if self._artifacts_client:
            return self._artifacts_client

        self._artifacts_client = artifacts_pb2_grpc.ArtifactRegistryStub(self.client._channel)
        return self._artifacts_client

    def get_artifact(
        self,
        uri: typing.Optional[str] = None,
        artifact_key: typing.Optional[art_id.ArtifactKey] = None,
        artifact_id: typing.Optional[art_id.ArtifactID] = None,
        query: typing.Optional[art_id.ArtifactQuery] = None,
        get_details: bool = False,
    ) -> typing.Optional[Artifact]:
        """
        Get the specified artifact.

        :param uri: An artifact URI.
        :param artifact_key: An artifact key.
        :param artifact_id: The artifact ID.
        :param query: An artifact query.
        :param get_details: A bool to indicate whether or not to return artifact details.
        :return: The artifact as persisted in the service.
        """
        if query:
            q = query
        elif uri:
            q = art_id.ArtifactQuery(uri=uri)
        elif artifact_key:
            q = art_id.ArtifactQuery(artifact_id=art_id.ArtifactID(artifact_key=artifact_key))
        elif artifact_id:
            q = art_id.ArtifactQuery(artifact_id=artifact_id)
        else:
            raise ValueError("One of uri, key, id")
        req = artifacts_pb2.GetArtifactRequest(query=q, details=get_details)
        resp = self.artifacts_client.GetArtifact(req)
        a = Artifact.from_flyte_idl(resp.artifact)
        if a.literal and a.name:
            # assigned here because the resolver has an implicit dependency on the remote's context
            # can move into artifact if we are okay just using the current context.
            a.resolver = LiteralsResolver(literals={a.name: a.literal}, variable_map=None, ctx=self.context)
        return a

    def _execute(
        self,
        entity: typing.Union[FlyteTask, FlyteWorkflow, FlyteLaunchPlan],
        inputs: typing.Dict[str, typing.Any],
        project: str = None,
        domain: str = None,
        execution_name: typing.Optional[str] = None,
        execution_name_prefix: typing.Optional[str] = None,
        options: typing.Optional[Options] = None,
        wait: bool = False,
        type_hints: typing.Optional[typing.Dict[str, typing.Type]] = None,
        overwrite_cache: typing.Optional[bool] = None,
        envs: typing.Optional[typing.Dict[str, str]] = None,
        tags: typing.Optional[typing.List[str]] = None,
        cluster_pool: typing.Optional[str] = None,
        **kwargs,
    ) -> FlyteWorkflowExecution:
        resolved_inputs = {}
        for k, v in inputs.items():
            # TODO: Remove when https://github.com/flyteorg/flytekit/pull/2136 gets merged
            if Artifact is None:
                resolved_inputs[k] = v
            elif isinstance(v, artifacts_pb2.Artifact):
                lit = v.spec.value
                resolved_inputs[k] = lit
            elif isinstance(v, Artifact):
                if v.literal is not None:
                    lit = v.literal
                elif v.to_id_idl() is not None:
                    fetched_artifact = self.get_artifact(artifact_id=v.concrete_artifact_id)
                    if not fetched_artifact:
                        raise user_exceptions.FlyteValueException(
                            v.concrete_artifact_id, "Could not find artifact with ID given"
                        )
                    lit = fetched_artifact.literal
                else:
                    raise user_exceptions.FlyteValueException(
                        v, "When binding input to Artifact, either the Literal or the ID must be set"
                    )
                resolved_inputs[k] = lit
            else:
                resolved_inputs[k] = v

        return super()._execute(
            entity,
            resolved_inputs,
            project=project,
            domain=domain,
            execution_name=execution_name,
            execution_name_prefix=execution_name_prefix,
            options=options,
            wait=wait,
            type_hints=type_hints,
            overwrite_cache=overwrite_cache,
            envs=envs,
            tags=tags,
            cluster_pool=cluster_pool,
            **kwargs,
        )

    def create_artifact(self, artifact: Artifact):
        """
        Create an artifact in FlyteAdmin.

        :param artifact: The artifact to create.
        :return: The artifact as persisted in the service.
        """
        # Two things can happen here -
        #  - the call to to_literal may upload something, in the case of an offloaded data type.
        #    - if this happens, the upload request should already return the created Artifact object.
        if artifact.literal is None:
            with self.remote_context() as ctx:
                lt = artifact.literal_type or TypeEngine.to_literal_type(artifact.python_type)
                lit = TypeEngine.to_literal(ctx, artifact.python_val, artifact.python_type, lt)
                artifact.literal_type = lt.to_flyte_idl()
                artifact.literal = lit.to_flyte_idl()
        else:
            if artifact.literal_type is None:
                raise ValueError("Cannot create an artifact without a literal type set.")

        if artifact.project is None:
            artifact.project = self.default_project
        if artifact.domain is None:
            artifact.domain = self.default_domain
        create_request = Artifact.to_create_request(artifact)
        logging.debug(f"CreateArtifact request {create_request}")
        resp = self.artifacts_client.CreateArtifact(create_request)
        logging.debug(f"CreateArtifact response {resp}")

    def _has_file_extension(self, file_path) -> bool:
        _, ext = os.path.splitext(file_path)
        return bool(ext)

    def get(self, uri: typing.Optional[str] = None) -> typing.Optional[typing.Union[LiteralsResolver, Literal, bytes]]:
        if not uri.startswith("union://"):
            # Assume this is the default behavior and this function is being called with a flyte uri
            return super().get(uri)
        if self._has_file_extension(uri):
            blob_dimensionality = BlobType.BlobDimensionality.SINGLE
        else:
            blob_dimensionality = BlobType.BlobDimensionality.MULTIPART
        return Literal(
            scalar=Scalar(
                blob=Blob(
                    metadata=BlobMetadata(
                        type=BlobType(
                            format="",
                            dimensionality=blob_dimensionality,
                        ),
                    ),
                    uri=uri,
                )
            )
        )

    @property
    def images_client(self) -> image_service_pb2_grpc.ImageServiceStub:
        if self._images_client:
            return self._images_client

        self._images_client = image_service_pb2_grpc.ImageServiceStub(self.client._channel)
        return self._images_client

    def _get_image_fqin(self, name: str) -> str:
        image_id = image_definition__pb2.ImageIdentifier(name=name)
        org = _get_organization(self.config.platform, channel=self.client._channel)
        req = image_payload__pb2.GetImageRequest(id=image_id, organization=org)
        resp = self.images_client.GetImage(req)
        return resp.image.fqin



# From https://github.com/python/cpython/blob/fb0cf7d1408c904e40142a74cd7a53eb52a8e568/Lib/functools.py#L444-L489

# Python module wrapper for _functools C module
# to allow utilities written in Python to be added
# to the functools module.
# Written by Nick Coghlan <ncoghlan at gmail.com>,
# Raymond Hettinger <python at rcn.com>,
# and Łukasz Langa <lukasz at langa.pl>.
#   Copyright (C) 2006-2013 Python Software Foundation.
# See C source code for _functools credits/copyright
from flytekit.types.directory import FlyteDirectory
from flytekit.types.file import FlyteFile


def _extract_key(obj):
    if isinstance(obj, (FlyteFile, FlyteDirectory)):
        if obj.remote_source:
            return obj.remote_source
        else:
            return obj.path
    return obj


class _HashedSeq(list):
    __slots__ = "hashvalue"

    def __init__(self, tup, hash=hash):
        self[:] = tup
        self.hashvalue = hash(tup)

    def __hash__(self):
        return self.hashvalue


def _make_key(args, kwds, kwd_mark=(object(),), fasttypes={int, str}, type=type, len=len):
    key = args
    if kwds:
        key += kwd_mark
        for item in kwds.items():
            key += item
    if len(key) == 1 and type(key[0]) in fasttypes:
        return key[0]
    key_ = tuple(_extract_key(k) for k in key)
    return _HashedSeq(key_)



import rich_click as click

from union.cli._app_common import DeployApplicationGroupForFiles


@click.group(name="deploy")
def deploy():
    """Deploy a resource."""


app_help = """Deploy application on Union."""
app_group = DeployApplicationGroupForFiles(
    name="apps",
    help=app_help,
    command_name="deploy",
)
deploy.add_command(app_group)



import asyncio
import atexit
import functools
import os
import threading
from contextlib import contextmanager
from typing import Any, Awaitable, Callable, TypeVar

from flytekit.loggers import logger
from typing_extensions import ParamSpec

T = TypeVar("T")
P = ParamSpec("P")


@contextmanager
def _selector_policy():
    original_policy = asyncio.get_event_loop_policy()
    try:
        if os.name == "nt" and hasattr(asyncio, "WindowsSelectorEventLoopPolicy"):
            asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

        yield
    finally:
        asyncio.set_event_loop_policy(original_policy)


class _TaskRunner:
    """A task runner that runs an asyncio event loop on a background thread."""

    def __init__(self) -> None:
        self.__loop: asyncio.AbstractEventLoop | None = None
        self.__runner_thread: threading.Thread | None = None
        self.__lock = threading.Lock()
        atexit.register(self._close)

    def _close(self) -> None:
        if self.__loop:
            self.__loop.stop()

    def _execute(self) -> None:
        loop = self.__loop
        assert loop is not None
        try:
            loop.run_forever()
        finally:
            loop.close()

    def run(self, coro: Any) -> Any:
        """Synchronously run a coroutine on a background thread."""
        name = f"{threading.current_thread().name} : loop-runner"
        with self.__lock:
            if self.__loop is None:
                with _selector_policy():
                    self.__loop = asyncio.new_event_loop()
                self.__runner_thread = threading.Thread(target=self._execute, daemon=True, name=name)
                self.__runner_thread.start()
        fut = asyncio.run_coroutine_threadsafe(coro, self.__loop)
        res = fut.result(None)

        return res


class _AsyncLoopManager:
    def __init__(self):
        self._runner_map: dict[str, _TaskRunner] = {}

    def run_sync(self, coro_func: Callable[..., Awaitable[T]], *args, **kwargs) -> T:
        """
        This should be called from synchronous functions to run an async function.
        """
        name = threading.current_thread().name + f"PID:{os.getpid()}"
        coro = coro_func(*args, **kwargs)
        if name not in self._runner_map:
            if len(self._runner_map) > 500:
                logger.warning(
                    "More than 500 event loop runners created!!! This could be a case of runaway recursion..."
                )
            self._runner_map[name] = _TaskRunner()
        return self._runner_map[name].run(coro)

    def synced(self, coro_func: Callable[P, Awaitable[T]]) -> Callable[P, T]:
        """Make loop run coroutine until it returns. Runs in other thread"""

        @functools.wraps(coro_func)
        def wrapped(*args: Any, **kwargs: Any) -> T:
            return self.run_sync(coro_func, *args, **kwargs)

        return wrapped


def exception_handler(loop: asyncio.AbstractEventLoop, context: dict) -> None:
    """Used by patch_exception_handler to patch BlockingIOError."""
    if (
        "exception" in context
        and type(context["exception"]).__name__ == "BlockingIOError"
        and str(context["exception"]).startswith("[Errno")
    ):
        return
    loop.default_exception_handler(context)


def patch_exception_handler(loop: asyncio.AbstractEventLoop) -> None:
    """Patch exception handler to ignore the `BlockingIOError: [Errno 11] ...` error.

    This is emitted by `aio.grpc` when multiple event loops are used in separate threads.
    This is an issue with grpc's cython implementation of `aio.Channel.__init__` where
    `socket.recv(1)` call only works on the first call. All subsequent calls result in
    an error, but this does not have any impact.

    For more info:
        - https://github.com/grpc/grpc/issues/25364
        - https://github.com/grpc/grpc/pull/36096
    """
    loop.set_exception_handler(exception_handler)


loop_manager = _AsyncLoopManager()
run_sync = loop_manager.run_sync
